import torch
from torch import nn

from wild_relation_network import relation_network
from wild_relation_network.layers import ConvBNReLU, GroupObjectsIntoPairs, GroupObjectsIntoPairsWith, TagPanelEmbeddings, Identity, \
    GroupObjectsIntoTriples, GroupObjectsIntoTriplesWith


class WReN(nn.Module):
    """
    Wild Relation Network (WReN) [1] for solving Raven's Progressive Matrices.
    The originally proposed model uses a Relation Network (RN) [2] which works on object pairs.
    This implementation allows to extend the RN to work on object triples (by setting use_object_triples=True).
    Using larger tuples is impractical, as the memory requirement grows exponentially,
    with complexity O(num_objects ^ rn_tuple_size).
    After extension to triples, the model resembles the Logic Embedding Network (LEN) [3].

    [1] Santoro, Adam, et al. "Measuring abstract reasoning in neural networks." ICML 2018
    [2] Santoro, Adam, et al. "A simple neural network module for relational reasoning." NeurIPS 2017
    [3] Zheng, Kecheng, Zheng-Jun Zha, and Wei Wei. "Abstract reasoning with distracting features." NeurIPS 2019
    """

    def __init__(self, num_channels: int = 32, use_object_triples: bool = False, use_layer_norm: bool = False):
        """
        Initializes the WReN model.
        :param num_channels: number of convolutional kernels in each CNN layer
        :param use_object_triples: flag indicating whether to group objects into triples (by default object pairs are considered)
        in the Relation Network submodule. Use False to reproduce WReN and True to reproduce LEN.
        :param use_layer_norm: flag indicating whether layer normalization should be applied after
        the G submodule of RN.
        """
        super(WReN, self).__init__()
        if use_object_triples:
            self.group_objects = GroupObjectsIntoTriples(num_objects=8)
            self.group_objects_with = GroupObjectsIntoTriplesWith()
        else:
            self.group_objects = GroupObjectsIntoPairs(num_objects=8)
            self.group_objects_with = GroupObjectsIntoPairsWith()

        self.cnn = nn.Sequential(
            ConvBNReLU(1, num_channels, kernel_size=3, stride=2),
            ConvBNReLU(num_channels, num_channels, kernel_size=3, stride=2),
            ConvBNReLU(num_channels, num_channels, kernel_size=3, stride=2),
            ConvBNReLU(num_channels, num_channels, kernel_size=3, stride=2)
        )
        self.object_size = num_channels * 9 * 9
        self.object_tuple_size = (3 if use_object_triples else 2) * (self.object_size + 9)
        self.tag_panel_embeddings = TagPanelEmbeddings()
        self.g = relation_network.G(
            depth=3,
            in_size=self.object_tuple_size,
            out_size=self.object_tuple_size,
            use_layer_norm=False
        )
        self.norm = nn.LayerNorm(self.object_tuple_size) if use_layer_norm else Identity()
        self.f = relation_network.F(
            depth=2,
            object_size=self.object_tuple_size,
            out_size=1
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the WReN model.
        :param x: a tensor with shape (batch_size, num_panels, height, width). num_panels is assumed
        to be 16, as the model was designed to solve RPMs from the PGM dataset.
        :return: a tensor with shape (batch_size, num_answers). num_answers is always equal to 8,
        which is the number of answers for each RPM in PGM.
        """
        batch_size, num_panels, height, width = x.size()
        x = x.view(batch_size * num_panels, 1, height, width)
        x = self.cnn(x)
        x = x.view(batch_size, num_panels, self.object_size)
        x = self.tag_panel_embeddings(x)
        context_objects = x[:, :8, :]
        choice_objects = x[:, 8:, :]
        context_pairs = self.group_objects(context_objects)
        context_g_out = self.g(context_pairs)
        f_out = torch.zeros(batch_size, 8, device=x.device).type_as(x)
        for i in range(8):
            context_choice_pairs = self.group_objects_with(context_objects, choice_objects[:, i, :])
            context_choice_g_out = self.g(context_choice_pairs)
            relations = context_g_out + context_choice_g_out
            relations = self.norm(relations)
            f_out[:, i] = self.f(relations).squeeze()
        return f_out



from typing import List

import torch
from torch import nn


class Identity(nn.Module):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x


class LinearBNReLU(nn.Module):
    def __init__(self, in_dim: int, out_dim: int):
        super(LinearBNReLU, self).__init__()
        self.linear = nn.Linear(in_dim, out_dim)
        self.bn = nn.BatchNorm1d(out_dim)
        self.relu = nn.ReLU()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.linear(x)
        shape = x.shape
        x = x.flatten(0, -2)
        x = self.bn(x)
        x = x.view(shape)
        x = self.relu(x)
        return x


class DeepLinearBNReLU(nn.Module):
    def __init__(self, depth: int, in_dim: int, out_dim: int, change_dim_first: bool = True):
        super(DeepLinearBNReLU, self).__init__()
        layers = []
        if change_dim_first:
            layers += [LinearBNReLU(in_dim, out_dim)]
            for _ in range(depth - 1):
                layers += [LinearBNReLU(out_dim, out_dim)]
        else:
            for _ in range(depth - 1):
                layers += [LinearBNReLU(in_dim, in_dim)]
            layers += [LinearBNReLU(in_dim, out_dim)]
        self.layers = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.layers(x)


class MLP(nn.Module):
    def __init__(self, in_dim: int, hidden_dims: List[int], out_dim: int):
        super(MLP, self).__init__()
        self.layers = nn.Sequential(*[
            LinearBNReLU(d1, d2)
            for d1, d2 in zip([in_dim] + hidden_dims, hidden_dims + [out_dim])
        ])

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.layers(x)


class ConvBNReLU(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, **kwargs):
        super(ConvBNReLU, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, **kwargs)
        self.bn = nn.BatchNorm2d(out_channels)
        self.relu = nn.ReLU()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv(x)
        x = self.bn(x)
        return self.relu(x)


class GroupObjectsIntoPairs(nn.Module):
    def __init__(self, num_objects: int):
        super().__init__()
        self.num_objects = num_objects

    def forward(self, objects: torch.Tensor) -> torch.Tensor:
        batch_size, num_objects, object_size = objects.size()
        return torch.cat([
            objects.unsqueeze(1).repeat(1, num_objects, 1, 1),
            objects.unsqueeze(2).repeat(1, 1, num_objects, 1)
        ], dim=3).view(batch_size, num_objects ** 2, 2 * object_size)

    def num_pairs(self):
        return self.num_objects ** 2


class GroupObjectsIntoPairsWith(nn.Module):
    def forward(self, objects: torch.Tensor, object: torch.Tensor) -> torch.Tensor:
        return torch.cat([
            objects,
            object.unsqueeze(1).repeat(1, 8, 1)
        ], dim=2)


class GroupObjectsIntoTriples(nn.Module):
    def __init__(self, num_objects: int):
        super().__init__()
        self.num_objects = num_objects

    def forward(self, objects: torch.Tensor) -> torch.Tensor:
        batch_size, num_objects, object_size = objects.size()
        return torch.cat([
            objects.unsqueeze(1).repeat(1, num_objects, 1, 1).unsqueeze(2).repeat(1, 1, num_objects, 1, 1),
            objects.unsqueeze(1).repeat(1, num_objects, 1, 1).unsqueeze(3).repeat(1, 1, 1, num_objects, 1),
            objects.unsqueeze(2).repeat(1, 1, num_objects, 1).unsqueeze(3).repeat(1, 1, 1, num_objects, 1)
        ], dim=4).view(batch_size, num_objects ** 3, 3 * object_size)

    def num_triples(self) -> int:
        return self.num_objects ** 3


class GroupObjectsIntoTriplesWith(nn.Module):
    def forward(self, objects: torch.Tensor, object: torch.Tensor) -> torch.Tensor:
        batch_size, num_objects, object_size = objects.size()
        return torch.cat([
            objects.unsqueeze(1).repeat(1, num_objects, 1, 1),
            objects.unsqueeze(2).repeat(1, 1, num_objects, 1),
            object.unsqueeze(1).unsqueeze(2).repeat(1, num_objects, num_objects, 1)
        ], dim=3).view(batch_size, num_objects ** 2, 3 * object_size)


class TagPanelEmbeddings(nn.Module):
    """ Tags panel embeddings of 3x3 Raven's Progressive Matrices (RPMs) with their absolute coordinates. """

    def forward(self, panel_embeddings: torch.Tensor) -> torch.Tensor:
        """
        Concatenates a one-hot encoded vector to each RPM panel.
        The concatenated vector indicates panel absolute position in the RPM.
        :param panel_embeddings: a tensor of shape (batch_size, 16, embedding_size)
        :return: a tensor of shape (batch_size, 16, embedding_size + 9)
        """
        batch_size = panel_embeddings.shape[0]
        tags = torch.zeros((16, 9), device=panel_embeddings.device).type_as(panel_embeddings)
        tags[:8, :8] = torch.eye(8, device=panel_embeddings.device).type_as(panel_embeddings)
        tags[8:, 8] = torch.ones(8, device=panel_embeddings.device).type_as(panel_embeddings)
        tags = tags.expand((batch_size, -1, -1))
        return torch.cat([panel_embeddings, tags], dim=2)



from setuptools import find_packages, setup

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='wild_relation_network',
    version='0.1.0',
    author='Mikołaj Małkiński',
    author_email='mikolaj.malkinski@gmail.com',
    license='MIT',
    description="Wild Relation Network for solving Raven's Progressive Matrices",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/mikomel/wild-relation-network',
    keywords=['artificial intelligence', 'relation network', 'wild relation network'],
    install_requires=[
        'torch>=1.5'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    packages=find_packages(),
    python_requires='>=3.6',
)



import torch
from torch import nn

from wild_relation_network.layers import GroupObjectsIntoPairs, DeepLinearBNReLU, Identity, GroupObjectsIntoTriples


class RelationNetwork(nn.Module):
    """
    Relation Network (RN) [1].
    The originally proposed model works on object pairs.
    This implementation allows to extend the RN to work on object triples (by setting tuple_size=3).
    Using larger tuples may be impractical, as the memory requirement grows exponentially,
    with complexity O(num_objects ^ tuple_size).

    [1] Santoro, Adam, et al. "A simple neural network module for relational reasoning." NeurIPS 2017
    """

    def __init__(self, num_objects: int, object_size: int, out_size: int, g_depth: int = 3,
                 f_depth: int = 2, use_object_triples: bool = False, use_layer_norm: bool = False):
        """
        Initializes the RN model.
        :param num_objects: number of objects (objects are represented as 1D tensors)
        :param object_size: object size
        :param out_size: output size
        :param g_depth: number of MLP layers in G
        :param f_depth: number of MLP layers in F
        :param use_object_triples: flag indicating whether to group objects into triples (by default object pairs are considered)
        :param use_layer_norm: flag indicating whether layer normalization should be applied after G submodule
        """
        super(RelationNetwork, self).__init__()
        self.group_objects = GroupObjectsIntoTriples(num_objects) if use_object_triples else GroupObjectsIntoPairs(num_objects)
        object_tuple_size = (3 if use_object_triples else 2) * object_size
        self.g = G(depth=g_depth, in_size=object_tuple_size, out_size=object_tuple_size, use_layer_norm=use_layer_norm)
        self.f = F(depth=f_depth, object_size=object_tuple_size, out_size=out_size, dropout_probability=0.5)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the RN model.
        :param x: a tensor with shape (batch_size, num_objects, object_size)
        :return: a tensor with shape (batch_size, out_size)
        """
        x = self.group_objects(x)
        x = self.g(x)
        x = self.f(x)
        return x


class G(nn.Module):
    def __init__(self, depth: int, in_size: int, out_size: int, use_layer_norm: bool = False):
        super(G, self).__init__()
        self.mlp = DeepLinearBNReLU(depth, in_size, out_size, change_dim_first=False)
        self.norm = nn.LayerNorm(out_size) if use_layer_norm else Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.mlp(x)
        x = x.sum(dim=1)
        x = self.norm(x)
        return x


class F(nn.Module):
    def __init__(self, depth: int, object_size: int, out_size: int, dropout_probability: float = 0.5):
        super(F, self).__init__()
        self.mlp = nn.Sequential(
            DeepLinearBNReLU(depth, object_size, object_size),
            nn.Dropout(dropout_probability),
            nn.Linear(object_size, out_size)
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.mlp(x)
        return x
